﻿namespace FrnakovnikViaNoseGrowingAppleFactory
{
    // Item classes

    interface IEdible
    {
        public int Energy { get; }
        public void Eat();
    }

    // Fruits

    interface IFruit { }

    class Apple : IFruit, IEdible
    {
        public int Energy { get; init; }

        public virtual void Eat()
        {
            Console.WriteLine($"Apple eaten, player got {Energy} units of energy.");
        }
    }

    class GoldenApple : Apple
    {
        public override void Eat()
        {
            base.Eat();
            Console.WriteLine("Golden apple increased player's luck.");
        }
    }

    class NoseGrowingApple : Apple
    {
        public override void Eat()
        {
            base.Eat();
            Console.WriteLine("Player's nose got 100-times longer.");
        }
    }

    class Peach : IFruit, IEdible
    {
        public int Energy => 8;

        public void Eat()
        {
            Console.WriteLine($"Peach eaten, player got {Energy} units of energy.");
        }
    }

    // Fruit factories

    interface IFruitFactory
    {
        public IFruit Create();
    }

    class AppleFactory : IFruitFactory
    {
        private int _applesDispensed = 0;
        private int _appleEnergy;

        public AppleFactory(int appleEnergy)
        {
            _appleEnergy = appleEnergy;
        }

        public IFruit Create()
        {
            if ((_applesDispensed++) % 2 == 0)
            {
                return new GoldenApple { Energy = _appleEnergy };
            }
            else
            {
                return new Apple { Energy = _appleEnergy };
            }
        }
    }

    class NoseGrowingAppleFactory : IFruitFactory
    {
        public IFruit Create() => new NoseGrowingApple { Energy = 7 };
    }

    class PeachFactory : IFruitFactory
    {
        public IFruit Create() => new Peach();
    }

    // Tree

    class Tree
    {
        private IFruitFactory _factory;

        public Tree(IFruitFactory factory)
        {
            _factory = factory;
        }

        public List<IFruit> Harvest()
        {
            return new List<IFruit> {
                _factory.Create(),
                _factory.Create(),
                _factory.Create()
            };
        }
    }

    // Game logic

    internal class Program
    {
        static void Main(string[] args)
        {
            // Apple-tree
            var goodAppleFactory = new AppleFactory(10);
            var goodTree = new Tree(goodAppleFactory);
            HarvestAndEatAllFruits(goodTree);

            // Frnakovnik
            var noseGrowingAppleFactory = new NoseGrowingAppleFactory();
            var frnakovnik = new Tree(noseGrowingAppleFactory);
            HarvestAndEatAllFruits(frnakovnik);
        }

        static void HarvestAndEatAllFruits(Tree tree)
        {
            var fruits = tree.Harvest();
            foreach (IFruit fruit in fruits)
            {
                if (fruit is IEdible edible)
                {
                    Console.WriteLine("+++ Nice! Got edible fruit, trying to eat it ...");
                    edible.Eat();
                }
                else
                {
                    Console.WriteLine("+++ Damn! Got inedible fruit.");
                }
            }
            Console.WriteLine();
        }
    }
}