﻿namespace FruitDispensingStrategy
{
    // Item classes

    interface IEdible
    {
        public int Energy { get; }
        public void Eat();
    }

    // Fruits

    interface IFruit { }

    class Apple : IFruit, IEdible
    {
        public int Energy { get; init; }

        public virtual void Eat()
        {
            Console.WriteLine($"Apple eaten, player got {Energy} units of energy.");
        }
    }

    class GoldenApple : Apple
    {
        public override void Eat()
        {
            base.Eat();
            Console.WriteLine("Golden apple increased player's luck.");
        }
    }

    class NoseGrowingApple : Apple
    {
        public override void Eat()
        {
            base.Eat();
            Console.WriteLine("Player's nose got 100-times longer.");
        }
    }

    class Peach : IFruit, IEdible
    {
        public int Energy => 8;

        public void Eat()
        {
            Console.WriteLine($"Peach eaten, player got {Energy} units of energy.");
        }
    }

    // Fruit factories

    interface IFruitFactory
    {
        public IFruit Create();
    }

    abstract class ConfigurableAppleFactoryBase : IFruitFactory
    {
        public required int AppleEnergy { get; init; }

        public abstract IFruit Create();
    }

    class AppleFactory : ConfigurableAppleFactoryBase
    {
        public override IFruit Create() => new Apple { Energy = AppleEnergy };
    }

    class GoldenAppleFactory : ConfigurableAppleFactoryBase
    {
        public override IFruit Create() => new GoldenApple { Energy = AppleEnergy };
    }

    class NoseGrowingAppleFactory : IFruitFactory
    {
        public IFruit Create() => new NoseGrowingApple { Energy = 7 };
    }

    class PeachFactory : IFruitFactory
    {
        public IFruit Create() => new Peach();
    }

    // Dispensing strategies

    interface IFruitDispensingStrategy
    {
        public IFruit DispenseFruit();
    }

    class SingleFruitDispensingStrategy : IFruitDispensingStrategy
    {
        private IFruitFactory _factory;

        public SingleFruitDispensingStrategy(IFruitFactory factory)
        {
            _factory = factory;
        }

        public IFruit DispenseFruit()
        {
            return _factory.Create();
        }
    }

    class PeriodicSpecialFruitDispensingStrategy : IFruitDispensingStrategy
    {
        private IFruitFactory _regularFruitFactory;
        private IFruitFactory _specialFruitFactory;
        private int _specialFruitModulo;
        private int _fruitsDispensed = 0;

        public PeriodicSpecialFruitDispensingStrategy(IFruitFactory regularFruitFactory, IFruitFactory specialFruitFactory, int regularFruitCountPerOneSpecial)
        {
            if (regularFruitCountPerOneSpecial <= 0)
            {
                throw new ArgumentOutOfRangeException(nameof(regularFruitCountPerOneSpecial), "Must be greater or equal to 1.");
            }
            _specialFruitModulo = regularFruitCountPerOneSpecial + 1;

            _regularFruitFactory = regularFruitFactory;
            _specialFruitFactory = specialFruitFactory;
        }

        public IFruit DispenseFruit()
        {
            if ((_fruitsDispensed++) % _specialFruitModulo == 0)
            {
                return _specialFruitFactory.Create();
            }
            else
            {
                return _regularFruitFactory.Create();
            }
        }
    }

    // Tree

    class Tree
    {
        private IFruitDispensingStrategy _dispensingStrategy;

        public Tree(IFruitDispensingStrategy dispensingStrategy)
        {
            _dispensingStrategy = dispensingStrategy;
        }

        public List<IFruit> Harvest()
        {
            return new List<IFruit> {
                _dispensingStrategy.DispenseFruit(),
                _dispensingStrategy.DispenseFruit(),
                _dispensingStrategy.DispenseFruit()
            };
        }
    }

    // Game logic

    internal class Program
    {
        static void Main(string[] args)
        {
            // Apple-tree
            var appleTree = new Tree(
                new PeriodicSpecialFruitDispensingStrategy(
                    new AppleFactory { AppleEnergy = 10 },
                    new GoldenAppleFactory { AppleEnergy = 10 },
                    1
                )
            );
            HarvestAndEatAllFruits(appleTree);

            // Peach-tree
            var peachTree = new Tree(
                new SingleFruitDispensingStrategy(
                    new PeachFactory()
                )
            );
            HarvestAndEatAllFruits(peachTree);

            // Frnakovnik
            var frnakovnik = new Tree(
                new SingleFruitDispensingStrategy(
                    new NoseGrowingAppleFactory()
                )
            );
            HarvestAndEatAllFruits(frnakovnik);

            // occasionally Frnakovnik
            var occasionallyFrnakovnik = new Tree(
                new PeriodicSpecialFruitDispensingStrategy(
                    new AppleFactory { AppleEnergy = 10 },
                    new NoseGrowingAppleFactory(),
                    2
                )
            );
            HarvestAndEatAllFruits(occasionallyFrnakovnik);
        }

        static void HarvestAndEatAllFruits(Tree tree)
        {
            var fruits = tree.Harvest();
            foreach (IFruit fruit in fruits)
            {
                if (fruit is IEdible edible)
                {
                    Console.WriteLine("+++ Nice! Got edible fruit, trying to eat it ...");
                    edible.Eat();
                }
                else
                {
                    Console.WriteLine("+++ Damn! Got inedible fruit.");
                }
            }
            Console.WriteLine();
        }
    }
}